/******************************************************************************
*******************************************************************************
**                                                                           **
**      MOST for C Swing and Gopher - A file viewer                          **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: John E. Davis                                  **
**                                                                           **
**      Modifications primarily by Foteos Macrides                           **
**                                                                           **
**      Copyright 1991, John E. Davis                                        **
**                                                                           **
**      Terms of Copyright: See MAIN.C                                       **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
/*
 *  MOST:  line.c
 *
 *  Modified for C SWING and GOPHER compatibility by Foteos Macrides
 *                                                MACRIDES@SCI.WFEB.EDU
 */

#include <stdio.h>
#include "buffer.h"
#include "externs.h"
#include "display.h"
#include "line.h"
  
extern int COLUMN;
extern unsigned char *M_POS;
extern int M_FLAG;
extern int W_LINE;
extern int TAB_R;
extern int ESC_NUM;

#define MAX_LINE_LEN 512

/*
**  take 16 binary characters and put them in displayable form
*/
int ascii_format_line(char *buff, char *str)
{
    int i,ii,j,k,ch,di,flag;
    char num_str[4];

    ii = 0;
    di = 40;
    flag = 1;

    for (i = 35;i<=di;i++) str[i] = ' ';  /* a gap */
    
    for (j = 0; j < 4; j++)
      {
          
          for (k = 0; k < 4; k++)
            {
                i = 4 * j + k;
                ch = buff[i];
                if (MOST_V_OPT)
                  {
                      if (((ch < 32) && (ch >= 0)) || (ch == 127))
                        {
                            str[ii++] = '^';
                            if (ch == 127)
                              str[ii++] = '?';
                            else
                              str[ii++] = ch + '@';
                            str[i + di] = '.';
                        }
                      else if (ch < 0)
                        {
                            ch = ch + 256; 
                            sprintf(num_str,"%02X", ch);
                            str[ii++] = num_str[0];
                            str[ii++] = num_str[1];
                            str[i + di] = '.';
                        }
                      else
                        {
                            str[ii++] = ' ';
                            str[ii++] = ch;
                            str[i + di] = ch;
                        }
                  }
                else
                  {
                      if (ch < 32)
                        {
                            if (ch < 0) ch = ch + 256;
                            str[i + di] = '.';
                        }
                      else str[i+di] = ch;
                      
                      sprintf(num_str,"%02X", ch);
                      str[ii++] = num_str[0];
                      str[ii++] = num_str[1];
                  }
                
            } /* k */
          str[ii++] = ' ';
      }
    di += 16;
    str[di] = '\0';
    return(di);
}
    

/*
**  analyse ASCII line for bolding or underlining,
**  expansion of tabs, or selective indentation display
*/
int analyse_line(unsigned char *beg,
                 unsigned char *end, char *out, char *attributes)
{
    int test, ii, fold, ii_max, j, ok, esc_flag = 0;
    unsigned char *b,*e,*b1,*e1,ch,attr,at_esc = ' ';

    test = 0;
    ii = 0;
    ii_max = 0;
    fold = 0;
    if (MOST_W_OPT)
      {
          if (W_LINE)
            {
                esc_flag = 1;
                if      (ESC_NUM == 1) at_esc = 'b';
                else if (ESC_NUM == 4) at_esc = 'u';
                else if (ESC_NUM == 7) at_esc = 'r';
                else esc_flag = 0;
            }
          if (M_FLAG)
            {
                b = beg;
                e = end;
		b1 = beg_of_line1();
                e1 = end_of_line1();
            }
      }

    /*
    **  while character attributes need to be assigned
    */
    while(ch = *beg, (beg <= end) && (ch != '\n') && (ch != '\0'))
      {
          beg++;
          if (ii > ii_max) /* beyond previous high       */
            fold = 0;      /* so turn off bolding via ^M */     
          attr = ' '; /* turn off underlining or bolding attribute */

          /*
          **  set up bolding of line via ^M
          */
          if (!MOST_V_OPT && ch == '\015')
            {
                if (beg <= end)
                  {
                      fold = 1;
                      if (ii > ii_max) ii_max = ii - 1; /* ^M contributes nil */
                      if (out[0] == '\007' || out[0] == '\013' ||
                          out[0] == '\014')
                         ii = 1;
                      else
                         ii = 0;
                      if (MOST_W_OPT && W_LINE)
                        {
                           if (*b1 != '\007' && *b1 != '\013' && *b1 != '\014')
                                beg = M_POS - b1 + b + 1;
                           else
                                beg = M_POS - b1 + b;
                           if (beg > e1)
                             {
                                 beg = e1 + 1;
                             }
                        }
                  }
            }

          /*
          **  set up bolding or underlining of character if '\b' (^H)
          */
          else if (!MOST_V_OPT && ch == '\b' && ii != 0)
            {
                test = 1;
                ii--;
            }
          
          /*
          **  set up or turn off bolding, underlining or reversing
          **  invoked via an <ESC>[#m found within the line
          */
          else if (!MOST_V_OPT && ch == '\033' && *beg == '[' &&
                   (*(beg + 1) == '0' || *(beg + 1) == '1' ||
                    *(beg + 1) == '4' || *(beg + 1) == '7') &&
                    *(beg + 2) == 'm')
            {
                beg++;
                esc_flag = 1;
                if      (*beg == '1') at_esc = 'b';
                else if (*beg == '4') at_esc = 'u';
                else if (*beg == '7') at_esc = 'r';
                else
                  {
                      at_esc = ' ';
                      esc_flag = 0;
                  }
                beg += 2;
            }
          else if (!MOST_V_OPT && ch == '\033' && *beg == '[' &&
                   *(beg + 1) == 'm')
            {
                beg += 2;
                at_esc = ' ';
                esc_flag = 0;
            }

          /*
          **  expand tabs to 8 spaces if '\t' (^I)
          */
          else if (ch == '\t' && (!MOST_T_OPT || !MOST_V_OPT))
            {
                if (MOST_W_OPT && W_LINE)
                  {
                      if (ii == 0)
                        j = TAB_R;
                      else
                        j = 8 * (ii/8 + 1) - ii + TAB_R;
                  }
                else
                  j = 8 * (ii/8 + 1) - ii;  /* 8 column tabs */
                while(j--)
                  {
                      out[ii] = ' ';
                      attributes[ii++] = ' ';
                  }
            }
            
          /*
          **  assign bolding or underlining attributes
          **  invoked via ^M or '\b' (^H)
          */
          else if (test || fold)
            {
                test = 0;
                if (ch == out[ii])       /* overstrike, therefore bold */
                  attr = 'b';
                else if (out[ii] == '_') /* underline */
                  attr = 'u';

                if (fold && ch == ' ')
                  ii++;                  /* space, therefore no overstrike */
                else
                  {
                      attributes[ii] = attr; /* assign attribute */
                      out[ii++] = ch;
                  }
                
            }

          /*
          **  assign bolding, underlining or reversing attributes
          **  invoked via an <ESC>[#m found within the line
          */
          else if (esc_flag)
            {
                attributes[ii] = at_esc; /* assign attribute */
                out[ii++] = ch;
            }

          /*
          **  no bold, underline, or reverse attribute
          */
          else
            {
                out[ii] = ch;
                attributes[ii++] = attr;
            }

          /*
          **  handle overstriking of wrapped lines
          **  via ^M if it is in a later line
          */
          if (!MOST_V_OPT && MOST_W_OPT && M_FLAG &&
	      (beg <= M_POS && beg == end + 1))
            {
                if (!W_LINE)
                  {
                      beg = M_POS + 1;
                      end = M_POS - b1 + e + 1;
		      if (end > e1) end = e1;
                  }
                else
                  {
                      if (*b1 != '\007' && *b1 != '\013' && *b1 != '\014')
                           beg = M_POS - b1 + b + 1;
                      else
                           beg = M_POS - b1 + b;
                      if (beg > e1) beg = end + 1;
                      else
                        {
                            end = M_POS - b1 + e + 1;
                            if (end > e1) end = e1;
                        }
                  }
                fold = 1;
                ii_max = ii - 1;
                if (out[0] == '\007' || out[0] == '\013' ||
                    out[0] == '\014')
                     ii = 1;
                else
                     ii = 0;
            }
      } /* while */
    if (fold) ii = ii_max + 1;

    /*
    **  omit lines indented more than SELECTIVE_I_DISPLAY
    */
    if ((beg > end) && SELECTIVE_I_DISPLAY && !MOST_W_OPT)
      {
          ok = 1;
          if (*beg == '\n') beg++;
          while ((*beg <= ' ') && ok)
            {
                if (*beg != '\n') beg++;
                if (beg >= EOB) break;
                if ((*beg == '\n') ||
                    (apparant_distance(beg) >= SELECTIVE_I_DISPLAY))
                  {
                      ok = 0;
                  }
            }
          if (!ok)
            {
                ok = 3;
                while(ok--)
                  {
                      out[ii] = '.'; 
                      attributes[ii++] = ' ';
                  }
            }
      } /* SELECTIVE_I_DISPLAY */

    out[ii] = '\0';
    return(ii);
}


/*
**  output a formatted line to the display
*/
void output(char *line, int len, char *attr, char d_char)
{
    int
      i,ii,k, bold,b_len,n_len, s_len, u_len, r_len, work_len, j, mark, ok,
      quit, count, max_col, save, dollar;
    char ch, at, out[MAX_LINE_LEN];
    char *n_str, *b_str, *s_str, *u_str, *r_str, work[40];

    b_str = "\033[1m";     /* 'bold'      escape sequence */
    b_len = strlen(b_str);
    u_str = "\033[4m";     /* 'underline' escape sequence */
    u_len = strlen(u_str);
    r_str = "\033[7m";     /* 'reverse'   escape sequence */
    r_len = strlen(r_str);
    n_str = "\033[m";      /* 'normal'    escape sequence */
    n_len = strlen(n_str);
    s_str = "\033[%dC";    /* 'd'x'space' escape sequence */
    s_len = strlen(s_str);

    i = 0;
    ii = 0;
    bold = 0;
    quit = 0;
    ok = 0;  /* 1 if ok to start filling the out line */
    max_col = COLUMN + SCREEN_WIDTH - 1;
    count = 0;
    dollar = 0;

    /*
    **  while characters in the line still need to be counted
    */
    while (ch = line[i], (i <= len) && ch != '\0' && ch != '\n' && !quit)
      {
          /*
          **  increment count
          */
          count++;

          /*
          **  if count is in window, ok to start filling the out line
          **  if count exceeds window, break
          */
          if (count >= COLUMN) ok = 1;
          if (count > max_col) break;
          
          /*
          **  if not Verbose mode act on character attributes
          */
          if (!MOST_V_OPT)
            {
                /*
                **  does bold, underline, or reverse need to be turned on?
                */
                at = attr[i];
                if ((at != ' ') && (!bold || count == COLUMN))
                  {
                      bold = 1;
                      /* u, b and r have same length */
                      if (ok) for (j = 0; j < b_len; j++)
                        {
                            if (at == 'b')
                              out[ii++] = b_str[j];
                            else if (at == 'u')
                              out[ii++] = u_str[j];
                            else
                              out[ii++] = r_str[j];
                        }
                      
                  }

                /*
                **  does bold, underline, or reverse need to be turned off?
                */
                else if ((at == ' ') && bold)
                  {
                      bold = 0;
                      if (ok) for (j = 0; j < n_len; j++) out[ii++] = n_str[j];
                  }

                /*
                **  if it's not an always-explicit escape or
                **  control character, don't count it
                */
                if ((ch < ' ' || ch > 126) &&
                    (ch != '\007') && (ch != '\013') && (ch != '\014') &&
                    (ch != '\033'))
                  {
                      count--;
                  }
            } /* if not Verbose mode */
          
          /*
          **  if ch is a space or series of spaces
          **  put spaces into out line if ok, and
          **  assess whether dollar should be set
          */
          if (ch == ' ')
            {
                j = 0;
                mark = ii;
                /* we always make this loop once */
                while ((i <= len) && (line[i++] == ' '))
                  {
                      j++;
                      if (ok) out[ii++] = ' ';
                  }
                if (i > len) quit = 1;
                i = i - 2;
                save = count;
                count += j - 1;  /* counted one at top of while */
                
                if (!ok)
                  {
                      if (count >= COLUMN)
                        {
                            ok = 1;
                            j = count - COLUMN + 1;
                            k = j;
                            while (k--) out[ii++] = ' ';
                        }
                  }
                
                if (count > max_col)
                  {
                      quit = 1;
                      j = max_col - save;
                  }
                
                if (ok && (j > s_len) && (count < max_col)) 
                  {
                      ii = mark;
                      sprintf(work,s_str,j);
                      work_len = strlen(work);
                      for (j = 0; j < work_len; j++) out[ii++] = work[j];
                  }
                else if (count > max_col)
                  {
                      dollar = mark + j;
                  }
            } /* if ch is a space */
          
          /*
          **  else, if ok to fill line and ch is ^G, ^K, ^L, a Verbose
          **  control character, or an escape that was not followed by [#m,
	  **  display the control char as reverse video char
          */
          else if (ok && ((ch == '\007') ||
                          (ch == '\013') ||
                          (ch == '\014') ||
                          (ch == '\033') ||
                         ((ch < ' ' || ch > 126) && MOST_V_OPT)))
            {
                for (j=0; j < r_len; j++)
                  out[ii++] = r_str[j];
                if (ch == 127)
                  out[ii++] = '?';
                else
                  out[ii++] = ch + '@';
                if (count == max_col)
		  bold = 1;
		else
                  for (j=0; j < n_len; j++)
                    out[ii++] = n_str[j];
            }
          
          /*
          ** else if ok to fill line and 
          **   ( (ch is not a space) and
          **     (ch is not a Verbose control char) )
          ** put ch into out line
          */
          else if (ok)
            {
                out[ii++] = ch;
            }
                      
          i++;
      } /* while */
    
    if (dollar) 
      {
          bold = 1;
          ii = dollar;
          for (j=0; j < r_len; j++)
            out[ii++] = r_str[j];
          out[ii++] = d_char;
      }
    else if (ok && count > max_col)
      {
	  ii--;
          if (ok && bold)
            for (j=0; j < n_len; j++)
              out[ii++] = n_str[j];
          bold = 1;
          for (j=0; j < r_len; j++)
            out[ii++] = r_str[j];
          out[ii++] = d_char;
      }
    out[ii] = '\0';
    if (ii) fputs(out,stdout);
    
    if (ok && bold) fputs(n_str,stdout);
} /* output */


/*
**  have an ASCII or binary line prepared/analyzed
**  and then call the output function
*/
void most_display_line()
{
    unsigned char *beg, *end;
    int i, len, v, t;
    char buff[16];
    char the_line[MAX_LINE_LEN],  *p, the_attr[MAX_LINE_LEN], *line,*attr, ch;

    line = the_line;
    attr = the_attr;

    /* This needs to be fixed for files with really big lines */
    /*
    **  ASCII line
    */
    if (!MOST_B_OPT)
      {
          /*
          **  fetch beginning and end of line
          **  and set the right margin character
          */
          if(extract_line(&beg, &end) && MOST_W_OPT) ch = '\\'; else ch = '>';
          len = end - beg + 1;

          /*
          **  if line is too long
          **  set Verbose and display of tabs
          */
          if (len > MAX_LINE_LEN)
            {
                v = MOST_V_OPT;
                t = MOST_T_OPT;
                MOST_V_OPT = 1;
                MOST_T_OPT = 1;
                line = (char *) beg;
            }

          /*
          **  else have line analyzed for character display attributes
          */
          else len = analyse_line(beg, end, line, attr);
      }

    /*
    **  binary line
    */
    else
      {
          ch = '$';
          i = 0;
          beg = C_POS;
          end = C_POS + 16;
          if (end > EOB) end = EOB;
          while(beg < end) buff[i++] = (char) *beg++;
          while(i < 15) buff[i++] = 0;
          len = ascii_format_line(buff,the_line);
          i = 0; while(i<132) attr[i++] = ' ';
      }
    
    /*
    **  call the output function
    */
    output(line,len,attr, ch);

    /*
    **  restore Verbose and tab display options for too-long ASCII lines
    */
    if (len > MAX_LINE_LEN)
      {
          MOST_V_OPT = v;
          MOST_T_OPT = t;
      }
}


/*
**  given a position in a line, return apparant distance from bol
**  expanding tabs, etc... up to pos
*/
int apparant_distance(unsigned char *pos)
{
    int i;
    unsigned char *cur_pos, *save_pos, ch, *e;
    extern int W_LINE;
    extern int TAB_R;

    cur_pos = C_POS;
    save_pos = pos;
    C_POS = pos;
    pos = beg_of_line();
    e = end_of_line();
    C_POS = cur_pos;

    i = 1;
    while(ch = *pos, pos++ < save_pos)
      {
          /*
          **  ^H (underline/overstrike character)
          */
          if (!MOST_V_OPT && ch == '\b')
            {
                if (i > 1) i--;
            }

          /*
          **  ^M (overstrike line)
          */
          else if (!MOST_V_OPT && ch == '\015')
            {
                if (i > 1) i = 1;
            }

          /*
          **  underline, bold, or reverse via <ESC>[#m
          */
          else if (!MOST_V_OPT && ch == '\033' && pos <= e - 2 &&
                   *pos == '[' &&
                   (*(pos + 1) == '1' || *(pos + 1) == '4' ||
                    *(pos + 1) == '7' || *(pos + 1) == '0') &&
                   *(pos + 2) == 'm')
            i -= 3;
          else if (!MOST_V_OPT && ch == '\033' && pos <= e - 1 &&
                   *pos == '[' && *(pos + 1) == 'm')
            i -= 2;

          /*
          **  ^I (8 column tabs)
          */
          else if (ch == '\t' && (!MOST_T_OPT || !MOST_V_OPT))
            {
                if (MOST_W_OPT && W_LINE)
                  {
                      if (i == 1)
                        i += TAB_R;
                      else
                        i = 8 * ((i - 1)/8 + 1) + 1 + TAB_R;
                  }
                else
                  i = 8 * ((i - 1)/8 + 1) + 1;
            }

          /*
          **  other escape or control chars
          */
          else if (ch < ' ' || ch > 126)
            {
                /*
                **  ^G (bell) , ^K (vertical tab), ^L (linefeed) and
                **  ^[ (escape) not followed by [#m are always displayed
                **  (as reverse video ch),
                **  otherwise non-Verbose control chars have no width
                */
                if (MOST_V_OPT || ch == '\007' || ch == '\013' ||
                                  ch == '\014' || ch == '\033')
                  {
                      i++;
                  }
            }
          /*
          **  text character
          */
          else i++;
      }
    return (i);
}
